/**h* NFCTools/nfc_cards.c
 *
 * NAME
 *   nfc_cards.c
 *
 * DESCRIPTION
 *   Manipulation of NFC contactless cards (hardware implementation of tags)
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"
#include "nfc_card_i.h"

static const unsigned char ATR_INNOVISION_TOPAZ[] = { 0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x02, 0x00, 0x30, 0x00, 0x00, 0x00, 0x00, 0x5A };
static const unsigned char ATR_NXP_MIFARE_1K[]    = { 0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x6A };
static const unsigned char ATR_NXP_MIFARE_4K[]    = { 0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x69 };
static const unsigned char ATR_NXP_MIFARE_UL[]    = { 0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x68 };
static const unsigned char ATR_NXP_MIFARE_UL_C[]  = { 0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0x00, 0x3A, 0x00, 0x00, 0x00, 0x00, 0x51 };
static const unsigned char ATR_NXP_DESFIRE[]      = { 0x3B, 0x81, 0x80, 0x01, 0x80, 0x80 };
static const unsigned char ATR_NFC_EMULATOR[]     = { 0x3B, 0x80, 0x01, 0x81 };

BOOL nfc_card_recognize(nfc_card_t card)
{
  int i;

	if (card == NULL)
	  return FALSE;

	card->type = 0;

	if (nfc_verbose)
	{
    printf("\tATR: ");
		for (i=0; i<(int) card->atrlen; i++)
		  printf("%02X", card->atr[i]);
		printf("\n");
	}

#ifndef ONLY_DESFIRE
	if ((card->atrlen == sizeof(ATR_INNOVISION_TOPAZ)) && !memcmp(card->atr, ATR_INNOVISION_TOPAZ, sizeof(ATR_INNOVISION_TOPAZ)))
	{
	  if (!nfc_card_topaz_recognize(card))
		  return FALSE;

	} else
	if ((card->atrlen == sizeof(ATR_NXP_MIFARE_1K)) && !memcmp(card->atr, ATR_NXP_MIFARE_1K, sizeof(ATR_NXP_MIFARE_1K)))
	{
    if (nfc_verbose)
		  printf("\tthis is a Mifare Classic 1K chip\n");
		card->type = CARD_NXP_MIFARE_1K;
		card->capacity = 720;
	} else
	if ((card->atrlen == sizeof(ATR_NXP_MIFARE_4K)) && !memcmp(card->atr, ATR_NXP_MIFARE_4K, sizeof(ATR_NXP_MIFARE_4K)))
	{
    if (nfc_verbose)
		  printf("\tthis is a Mifare Classic 4K chip\n");
		card->type = CARD_NXP_MIFARE_4K;
		card->capacity = 3456;
	} else
	if (((card->atrlen == sizeof(ATR_NXP_MIFARE_UL)) && !memcmp(card->atr, ATR_NXP_MIFARE_UL, sizeof(ATR_NXP_MIFARE_UL)))
   || ((card->atrlen == sizeof(ATR_NXP_MIFARE_UL_C)) && !memcmp(card->atr, ATR_NXP_MIFARE_UL_C, sizeof(ATR_NXP_MIFARE_UL_C))))
	{
	  if (!nfc_card_ultralight_recognize(card))
		  return FALSE;
	} else
#endif
	if ((card->atrlen == sizeof(ATR_NXP_DESFIRE)) && !memcmp(card->atr, ATR_NXP_DESFIRE, sizeof(ATR_NXP_DESFIRE)))
	{
    if (nfc_verbose)
		  printf("\tthis is a chip in the ISO 14443-4 type A family\n\tmaybe this is a Desfire\n");
	  if (!nfc_card_desfire_recognize(card))
		  return FALSE;
	} else
	if ((card->atrlen == sizeof(ATR_NFC_EMULATOR)) && !memcmp(card->atr, ATR_NFC_EMULATOR, sizeof(ATR_NFC_EMULATOR)))
	{
    if (nfc_verbose)
		  printf("\tthis is a chip in the ISO 14443-4 type A family\n\tmaybe this is an NFC emulator\n");
	  if (!nfc_card_desfire_recognize(card))
		  return FALSE;
	} else
	{
    if (nfc_verbose)
		  printf("\tthis ATR is not registered\n");
	}

	return TRUE;
}

int  nfc_card_get_type(nfc_card_t card)
{
  if (card == NULL)
	  return 0;
	return card->type;
}

size_t nfc_card_get_capacity(nfc_card_t card)
{
  if (card == NULL)
	  return 0;
	return card->capacity;
}

BOOL nfc_card_is_prepared(nfc_card_t card)
{
	if (card == NULL)
	  return FALSE;

  return card->prepared;
}

BOOL nfc_card_is_writable(nfc_card_t card)
{
	if (card == NULL)
	  return FALSE;

  return card->writable;
}

BOOL nfc_card_is_empty(nfc_card_t card)
{
	if (card == NULL)
	  return FALSE;

  return card->empty;
}

BOOL nfc_card_unprepare(nfc_card_t card)
{
	if (card == NULL)
	  return FALSE;

  return TRUE;
}

BOOL nfc_card_prepare(nfc_card_t card)
{
	if (card == NULL)
	  return FALSE;

	switch (card->type)
	{
#ifndef ONLY_DESFIRE
    case CARD_INNOVISION_TOPAZ   : return nfc_card_topaz_prepare(card);

    case CARD_NXP_MIFARE_UL      : 
    case CARD_NXP_MIFARE_UL_C    : return nfc_card_ultralight_prepare(card);

    case CARD_NXP_MIFARE_1K      :
    case CARD_NXP_MIFARE_4K      : return nfc_card_mifare_prepare(card);

#endif
    case CARD_NXP_DESFIRE_EV1    : return nfc_card_desfire_prepare(card);

    default                      : return FALSE;
	}
}

BOOL nfc_card_lock(nfc_card_t card)
{
	if (card == NULL)
	  return FALSE;

	switch (card->type)
	{
#ifndef ONLY_DESFIRE
    case CARD_INNOVISION_TOPAZ   : return nfc_card_topaz_lock(card);

    case CARD_NXP_MIFARE_UL      : 
    case CARD_NXP_MIFARE_UL_C    : return nfc_card_ultralight_lock(card);

    case CARD_NXP_MIFARE_1K      :
    case CARD_NXP_MIFARE_4K      : return nfc_card_mifare_lock(card);
#endif

    case CARD_NXP_DESFIRE_EV1    : return nfc_card_desfire_lock(card);

    default                      : return FALSE;
	}
}

BOOL nfc_card_write_ndef(nfc_card_t card, unsigned char *ndef_buffer, size_t ndef_size)
{
  unsigned char *content;
	size_t physical_length, populated_length, content_length = 0;
  BOOL top_level_tlv = TRUE;
	BOOL result = FALSE;

	if (card == NULL)
	  return FALSE;

  if (card->type == CARD_NXP_DESFIRE_EV1)
    top_level_tlv = FALSE;

  if (top_level_tlv)
  {
	  if (ndef_size < 0xFF)
      populated_length = 2 + ndef_size;
	  else
      populated_length = 4 + ndef_size;
  } else
  {
    populated_length = ndef_size;
  }
		
	if (populated_length > card->capacity)
	{
	  if (!nfc_silent)
	    printf("Tag is too small for the NDEF\n");
	  return FALSE;
	}

	physical_length = populated_length;
	if (physical_length < card->capacity)
	  physical_length = card->capacity;

	content = nfc_alloc(physical_length);

  if (top_level_tlv)
  {
	  content[content_length++] = NDEF_MESSAGE_TAG;
	  if (ndef_size < 0xFF)
	  {
      content[content_length++] = (unsigned char) ndef_size;
	  } else
	  {
      content[content_length++] = 0xFF;
		  content[content_length++] = (unsigned char) (ndef_size / 0x0100);
		  content[content_length++] = (unsigned char) (ndef_size % 0x0100);
	  }
  }

	memcpy(&content[content_length], ndef_buffer, ndef_size);
  content_length += ndef_size;

  if (top_level_tlv)
  {
	  if (populated_length < physical_length)
      content[content_length++] = TERMINATOR_MESSAGE_TAG;
	  if (populated_length < physical_length)
		  content[content_length++] = 0;
  }

	switch (card->type)
	{
#ifndef ONLY_DESFIRE
    case CARD_INNOVISION_TOPAZ   : result = nfc_card_topaz_write_content(card, content, populated_length);
		                               break;

    case CARD_NXP_MIFARE_UL      : 
    case CARD_NXP_MIFARE_UL_C    : result = nfc_card_ultralight_write_content(card, content, populated_length);
		                               break;

    case CARD_NXP_MIFARE_1K      :
    case CARD_NXP_MIFARE_4K      : result = nfc_card_mifare_write_content(card, content, populated_length);
		                               break;
#endif

    case CARD_NXP_DESFIRE_EV1    : result = nfc_card_desfire_write_content(card, content, populated_length);
		                               break;

    default                      : break;
	}

  nfc_free(content);
	return result;
}
