/**h* NFCTools/nfc_desfire_format_main.c
 *
 * NAME
 *   nfc_desfire_format_main.c
 *
 * DESCRIPTION
 *   Main of nfc_desfire_format
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 15/02/2011 : first public release
 *
 **/
#include "nfc.h"
#include "cardware/desfire/pcsc_desfire.h"

#ifdef WIN32
#include "getopt/getopt.h"
#include "getopt/getopt_long.h"
#endif

#ifdef __linux__
#include <getopt.h>
#endif

static void hello(void);
static void usage(void);
static void helpmsg(const char *str);

BOOL nfc_verbose = FALSE;
BOOL nfc_silent  = FALSE;

const char prog_name[] = "nfc_desfire";
const char short_options[] = "epwz:k:t:r:bclyvqh";

const struct option long_options[] = {

   { "erase",          0, NULL, 'e' },
	 { "prepare",        0, NULL, 'p' },
	 { "lock",           0, NULL, 'w' },

   { "size",           1, NULL, 'z' },
   { "key",            1, NULL, 'k' },
	 { "key-type",       1, NULL, 't' },

   { "set-key",        1, NULL, 0x80 },
	 { "set-key-type",   1, NULL, 0x81 },

	 { "reader",         1, NULL, 'r' },
	 { "list-readers",   1, NULL, 'l' },

	 { "batch",          0, NULL, 'b' },
	 { "continue",       0, NULL, 'c' },

	 { "confirm",        0, NULL, 'y' },
	 { "verbose",        0, NULL, 'v' },
	 { "quiet",          0, NULL, 'q' },

	 { "help",           1, NULL, 'h' },

   { NULL,             0, NULL, 0 },

};


#define KEY_DES_OR_3DES     1
#define KEY_DES_OR_3DES_ISO 2
#define KEY_AES_ISO         3


int main(int argc, char **argv)
{
  int p_erase = 0;
	int p_prepare = 0;
	int p_lock = 0;
	int p_batch = 0;
	int p_continue = 0;
  char *p_key = NULL;
  int p_key_type = 0;
  char *p_set_key = NULL;
  int p_set_key_type = 0;
	char *p_reader = NULL;
  unsigned long p_size = 0;
	int p_confirm = 0;
	int p_list = 0;

	const unsigned char blank_buffer[32] = { 0 };

	unsigned char key_buffer[24];
	size_t key_length = 0;
	unsigned char set_key_buffer[24];
	size_t set_key_length = 0;

	int c;

	LONG rc;
	DF_VERSION_INFO version_info;
	DWORD ndef_size;
	DWORD cc_size = 32;

	nfc_card_t card = NULL;

  while ((c = getopt_long(argc, argv, short_options, long_options, NULL)) != -1)
  {
    switch (c)
    {
		  case 'e'  : p_erase = TRUE; break;
			case 'p'  : p_prepare = TRUE; break;
			case 'w'  : p_lock = TRUE; break;

			case 'z'  : p_size = atol(optarg); break;

			case 'k'  : p_key = optarg; break;
			case 't'  : p_key_type = atol(optarg); break;
			case 0x80 : p_set_key = optarg; break;
			case 0x81 : p_set_key_type = atol(optarg); break;

      case 'r'  : p_reader = optarg; break;
      case 'l'  : p_list = 1; break;

      case 'b'  : p_batch = 1; break;
      case 'c'  : p_continue = 1; break;

      case 'y'  : p_confirm = TRUE; break;
      case 'v'  : nfc_verbose = TRUE; break;
      case 'q'  : nfc_silent = TRUE; break;

      case 'h'  : hello();
                  usage();
                  return EXIT_FAILURE;

      case '?'  : hello();
			            helpmsg("Incorrect parameter(s).\n");
                  return EXIT_FAILURE;
									break;

      default   : hello();
			            helpmsg("Invalid command line.\n");
                  return EXIT_FAILURE;
    }
  }

	if (!nfc_silent)
	  hello();

  if (p_list)
	{
    nfc_list_readers();
		return EXIT_SUCCESS;
	}

  /* Check the minimum mandatory parameters */
	if (!(p_erase + p_prepare + p_lock))
	{
	  helpmsg("No mode specified, nothing to do!\n");
    return EXIT_FAILURE;
	}
	if ((p_erase + p_prepare) && (p_lock))
	{
	  helpmsg("Multiple modes specified, choose one!\n");
    return EXIT_FAILURE;
  }

  /* Retrieve the key(s) */
	if (p_erase || p_prepare)
	{
	  if (p_key == NULL)
		{
		  if (nfc_verbose)
			  printf("Key not specified, using default\n");

			memset(key_buffer, 0, sizeof(key_buffer));
      key_length = 8;
		} else
		{
		  if (!nfc_hex_to_bytes(key_buffer, sizeof(key_buffer), &key_length, p_key, FALSE))
			{
				helpmsg("Key: invalid parameter\n");
				return EXIT_FAILURE;
			}
		}

		if ((key_length  != 8) && (key_length  != 16) && (key_length  != 24))
		{
			helpmsg("Key: invalid parameter\n");
			return EXIT_FAILURE;
		}

		if (key_length == 8)
		{
		  memcpy(&key_buffer[8],  &key_buffer[0], 8);
			memcpy(&key_buffer[16], &key_buffer[0], 8);
		} else
		if (key_length == 16)
		{
		  memcpy(&key_buffer[16],  &key_buffer[0], 8);
		}

		if (p_key_type == 0)
		{
		  if (key_length == 8)
			{
				if (!nfc_silent)
					printf("Type of key not specified, assuming DES\n");
				p_key_type = KEY_DES_OR_3DES;
			} else
		  if (key_length == 16)
			{
				if (!nfc_silent)
					printf("Type of key not specified, assuming TripleDES\n");
				p_key_type = KEY_DES_OR_3DES;
			} else
		  if (key_length == 24)
			{
				if (!nfc_silent)
					printf("Type of key not specified, assuming TripleDES with 3 keys\n");
				p_key_type = KEY_DES_OR_3DES_ISO;
			} else
      {
				helpmsg("You must specify the type of the key!\n");
				return EXIT_FAILURE;
			}
		}

	  if (p_set_key != NULL)
		{
		  if (!nfc_hex_to_bytes(set_key_buffer, sizeof(set_key_buffer), &set_key_length, p_set_key, FALSE))
			{
				helpmsg("New key: invalid parameter\n");
				return EXIT_FAILURE;
			}

			if ((set_key_length  != 8) && (set_key_length  != 16) && (set_key_length  != 24))
			{
				helpmsg("New key: invalid parameter\n");
				return EXIT_FAILURE;
			}

			if (set_key_length == 8)
			{
				memcpy(&set_key_buffer[8],  &set_key_buffer[0], 8);
				memcpy(&set_key_buffer[16], &set_key_buffer[0], 8);
			} else
			if (set_key_length == 16)
			{
				memcpy(&set_key_buffer[16],  &set_key_buffer[0], 8);
			}

			if (p_set_key_type == 0)
			{
				if (set_key_length == 8)
				{
					if (!nfc_silent)
						printf("Type of new key not specified, assuming DES\n");
					p_set_key_type = KEY_DES_OR_3DES;
				} else
				if (set_key_length == 16)
				{
					if (!nfc_silent)
						printf("Type of new key not specified, assuming TripleDES\n");
					p_set_key_type = KEY_DES_OR_3DES;
				} else
				if (set_key_length == 24)
				{
					if (!nfc_silent)
						printf("Type of new key not specified, assuming TripleDES with 3 keys\n");
					p_set_key_type = KEY_DES_OR_3DES_ISO;
				} else
				{
					helpmsg("You must specify the type of the new key!\n");
					return EXIT_FAILURE;
				}
			}
		}
	}

  if (p_reader == NULL)
  {
    helpmsg("No Reader selected!\n");
    return EXIT_FAILURE;
  }

	/* Verify the reader */
	(const char *) p_reader = nfc_reader_find(p_reader);
	if (p_reader == NULL)
	{
    printf("Reader not found!\n");
    return EXIT_FAILURE;
	}

	if (!nfc_silent)
		printf("Looking for Desfire EV1 card on reader '%s'\n", p_reader);

batch_begin:

	if (card != NULL)
	{
    nfc_card_close(card);
    card = NULL;
	}

  /* Wait for a card */
  if (p_batch)
	{
		if (!nfc_silent)
			printf("Put next Desfire EV1 card on the reader...\n");
		if (!nfc_reader_wait_card_present(p_reader))
		{
			printf("Failed to wait for card arrival.\n");
			goto failed;
		}
	} else
	{
		if (!nfc_reader_is_card_present(p_reader))
		{
			printf("Put a Desfire EV1 card on the reader...\n");
			if (!nfc_reader_wait_card_present(p_reader))
			{
				printf("Failed to wait for card arrival.\n");
				goto failed;
			}
		}
	}

  /* Connect to the card */
  card = nfc_card_open(p_reader);
	if (card == NULL)
	{
    printf("Failed to connect to the target card.\n");
		if (p_batch && p_continue)
			goto batch_begin;
		goto failed;
	}

	/* Open the Desfire DLL */
	rc = SCardDesfire_AttachLibrary(nfc_card_handle(card));
  if (rc != SCARD_S_SUCCESS)
	{
    printf("Failed to instantiate the PC/SC Desfire DLL.\n");
    goto failed;
	}

	rc = SCardDesfire_IsoWrapping(nfc_card_handle(card), DF_ISO_WRAPPING_CARD);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("Failed to select the ISO 7816 wrapping mode.\n");
    goto failed;
	}

	rc = SCardDesfire_GetVersion(nfc_card_handle(card), &version_info);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("Desfire 'get version' command failed.\n");
		if (p_batch && p_continue)
			goto batch_end;
		goto failed;
	}

	if (!nfc_silent)
	  printf("Found a Desfire card\n");

  if (nfc_verbose)
	{
		printf("Hardware: Vendor=%02X, Type=%02X, SubType=%02X, v%02X.%02X\n", version_info.bHwVendorID,
		                                                                       version_info.bHwType,
		                                                                       version_info.bHwSubType,
		                                                                       version_info.bHwMajorVersion,
		                                                                       version_info.bHwMinorVersion);
		printf("Software: Vendor=%02X, Type=%02X, SubType=%02X, v%02X.%02X\n", version_info.bSwVendorID,
		                                                                       version_info.bSwType,
		                                                                       version_info.bSwSubType,
		                                                                       version_info.bSwMajorVersion,
		                                                                       version_info.bSwMinorVersion);
  }

	if ((version_info.bHwVendorID != 0x04) || (version_info.bSwVendorID != 0x04))
	{
	  printf("Manufacturer is not NXP\n");
		if (p_batch && p_continue)
			goto batch_end;
		goto failed;
	}

	if ((version_info.bHwType != 0x01) || (version_info.bSwType != 0x01))
	{
	  printf("Type is not Desfire\n");
		if (p_batch && p_continue)
			goto batch_end;
		goto failed;
	}

	if (version_info.bSwMajorVersion < 1)
	{
	  printf("Software version is below EV1\n");
		if (p_batch && p_continue)
			goto batch_end;
		goto failed;
	}

	if (p_erase || p_prepare)
	{
    if (nfc_verbose)
		  printf("Authenticating...\n");

	  switch (p_key_type)
		{
		  case KEY_DES_OR_3DES     : rc = SCardDesfire_Authenticate(nfc_card_handle(card), 0, key_buffer); break;
			case KEY_DES_OR_3DES_ISO : rc = SCardDesfire_AuthenticateIso24(nfc_card_handle(card), 0, key_buffer); break;
			case KEY_AES_ISO         : rc = SCardDesfire_AuthenticateAes(nfc_card_handle(card), 0, key_buffer); break;
		  default                  : rc = -1; break;
		}
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Authentication failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

		if (set_key_length)
		{
			if (nfc_verbose)
				printf("Setting the new key...\n");

			if ((p_key_type == KEY_DES_OR_3DES) && (p_set_key_type == KEY_DES_OR_3DES))
			{
        rc = SCardDesfire_ChangeKey(nfc_card_handle(card), 0, set_key_buffer, NULL);
			} else
			if ((p_key_type == KEY_DES_OR_3DES) && (p_set_key_type == KEY_DES_OR_3DES_ISO))
			{
        rc = SCardDesfire_AuthenticateIso24(nfc_card_handle(card), 0, key_buffer);
				if (rc == SCARD_S_SUCCESS)
				  rc = SCardDesfire_ChangeKey24(nfc_card_handle(card), 0, set_key_buffer, NULL);
			} else
			if ((p_key_type == KEY_DES_OR_3DES) && (p_set_key_type == KEY_AES_ISO))
			{
        rc = SCardDesfire_ChangeKeyAes(nfc_card_handle(card), DF_APPLSETTING2_AES | 0, 0, set_key_buffer, NULL);
			} else
			if ((p_key_type == KEY_DES_OR_3DES_ISO) && (p_set_key_type == KEY_DES_OR_3DES))
			{
			  rc = SCardDesfire_ChangeKeyAes(nfc_card_handle(card), DF_APPLSETTING2_AES | 0, 0, blank_buffer, NULL);
				if (rc == SCARD_S_SUCCESS)
				  rc = SCardDesfire_AuthenticateAes(nfc_card_handle(card), 0, blank_buffer);
				if (rc == SCARD_S_SUCCESS)
				  rc = SCardDesfire_ChangeKey(nfc_card_handle(card), 0, set_key_buffer, NULL);
			} else
			if ((p_key_type == KEY_DES_OR_3DES_ISO) && (p_set_key_type == KEY_DES_OR_3DES_ISO))
			{
			  rc = SCardDesfire_ChangeKey24(nfc_card_handle(card), 0, set_key_buffer, NULL);
			} else
			if ((p_key_type == KEY_DES_OR_3DES_ISO) && (p_set_key_type == KEY_AES_ISO))
			{
        rc = SCardDesfire_ChangeKeyAes(nfc_card_handle(card), DF_APPLSETTING2_AES | 0, 0, set_key_buffer, NULL);
			} else
			if ((p_key_type == KEY_AES_ISO) && (p_set_key_type == KEY_DES_OR_3DES))
			{
        rc = SCardDesfire_ChangeKey(nfc_card_handle(card), 0, set_key_buffer, NULL);
			} else
			if ((p_key_type == KEY_AES_ISO) && (p_set_key_type == KEY_DES_OR_3DES_ISO))
			{
        rc = SCardDesfire_ChangeKey24(nfc_card_handle(card), 0, set_key_buffer, NULL);
			} else
			if ((p_key_type == KEY_AES_ISO) && (p_set_key_type == KEY_AES_ISO))
			{
        rc = SCardDesfire_ChangeKeyAes(nfc_card_handle(card), DF_APPLSETTING2_AES | 0, 0, set_key_buffer, NULL);
			}

			if (rc != SCARD_S_SUCCESS)
			{
				printf("Change key failed\n");
				if (!nfc_silent)
					printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
				if (p_batch && p_continue)
					goto batch_end;
				goto failed;
			}

			switch (p_set_key_type)
			{
				case KEY_DES_OR_3DES     : rc = SCardDesfire_Authenticate(nfc_card_handle(card), 0, set_key_buffer); break;
				case KEY_DES_OR_3DES_ISO : rc = SCardDesfire_AuthenticateIso24(nfc_card_handle(card), 0, set_key_buffer); break;
				case KEY_AES_ISO         : rc = SCardDesfire_AuthenticateAes(nfc_card_handle(card), 0, set_key_buffer); break;
				default                  : rc = -1; break;
			}
			if (rc != SCARD_S_SUCCESS)
			{
				printf("Authentication with new key failed\n");
				if (!nfc_silent)
					printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
				if (p_batch && p_continue)
					goto batch_end;
				goto failed;
			}

		}
	}

	if (p_erase)
	{
    if (nfc_verbose)
		  printf("Formating the card...\n");

		rc = SCardDesfire_FormatPICC(nfc_card_handle(card));
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Format PICC failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

    if (p_set_key != NULL)
		{
			if (nfc_verbose)
				printf("Changing the key...\n");

		}

    if (!nfc_silent)
		  printf("Card erase done...\n");
	}

	if (p_prepare)
	{
	  const unsigned char nfc_aid[] = NFC_NDEF_7816_AID;
    unsigned char cc_buffer[15];

    if (nfc_verbose)
		  printf("Creating the NFC NDEF application...\n");

		rc = SCardDesfire_CreateIsoApplication(nfc_card_handle(card),
		                                       NFC_NDEF_DESFIRE_DF_ID,
																				   0xFF,
																				   DF_APPLSETTING2_ISO_EF_IDS | 0,
																					 NFC_NDEF_7816_DF_ID,
																					 nfc_aid,
																					 sizeof(nfc_aid));
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Create application failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

    if (nfc_verbose)
		  printf("Entering the NFC NDEF directory...\n");

		rc = SCardDesfire_SelectApplication(nfc_card_handle(card), NFC_NDEF_DESFIRE_DF_ID);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Select application failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

    if (nfc_verbose)
		  printf("Creating the NFC CC file...\n");

		rc = SCardDesfire_CreateIsoStdDataFile(nfc_card_handle(card),
		                                       NFC_NDEF_DESFIRE_CC_EF_ID,
																					 NFC_NDEF_7816_CC_EF_ID,
																				   0,
																			     0xEEEE,
																				   cc_size);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Create CC file failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

		rc = SCardDesfire_GetFreeMemory(nfc_card_handle(card), &ndef_size);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Get free memory failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

		if (p_size != 0)
		{
		  if (ndef_size < p_size)
			{
				printf("NDEF size=%d, bigger than available space on card (%d)\n", p_size, ndef_size);
				if (!nfc_silent)
					printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
				if (p_batch && p_continue)
					goto batch_end;
				goto failed;
			}
			ndef_size = p_size;
		} 

    if (nfc_verbose)
		  printf("Creating the NFC NDEF file...\n");

		rc = SCardDesfire_CreateIsoStdDataFile(nfc_card_handle(card),
		                                       NFC_NDEF_DESFIRE_DATA_EF_ID,
																				   NFC_NDEF_7816_DATA_EF_ID,
															 				     0,
																				   0xEEEE,
																				   ndef_size);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Create NDEF file failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

    if (nfc_verbose)
		  printf("Populating the NFC CC file...\n");

		cc_buffer[0]  = 0x00;
		cc_buffer[1]  = 0x0F;
		cc_buffer[2]  = NFC_NDEF_VERSION_2;
		cc_buffer[3]  = 0x00;
		cc_buffer[4]  = 0x3B;
		cc_buffer[5]  = 0x00;
		cc_buffer[6]  = 0x34;
		cc_buffer[7]  = NDEF_FILE_CONTROL_TAG;
		cc_buffer[8]  = 0x06;
		cc_buffer[9]  = (unsigned char) (NFC_NDEF_7816_DATA_EF_ID >> 8);
		cc_buffer[10] = (unsigned char) (NFC_NDEF_7816_DATA_EF_ID);
		cc_buffer[11] = (unsigned char) (ndef_size >> 8);
		cc_buffer[12] = (unsigned char) (ndef_size);
		cc_buffer[13] = 0x00;
		cc_buffer[14] = 0x00;

    rc = SCardDesfire_WriteData2(nfc_card_handle(card),
                                 NFC_NDEF_DESFIRE_CC_EF_ID,
                                 0,
                                 sizeof(cc_buffer),
                                 cc_buffer);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Populate CC file failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

    if (!nfc_silent)
		  printf("Card prepare done...\n");
	}

	if (p_lock)
	{
	  unsigned char cc_lock = 0xFF;

    if (nfc_verbose)
		  printf("Entering the NFC NDEF directory...\n");

		rc = SCardDesfire_SelectApplication(nfc_card_handle(card), NFC_NDEF_DESFIRE_DF_ID);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Select application failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

		if (nfc_verbose)
		  printf("Locking the NDEF file\n");

	  rc = SCardDesfire_ChangeFileSettings(nfc_card_handle(card),
		                                     NFC_NDEF_DESFIRE_DATA_EF_ID,
																				 0,
																				 0xEFFF);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Change file settings failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

		if (nfc_verbose)
		  printf("Updating the CC file\n");

    rc = SCardDesfire_WriteData2(nfc_card_handle(card),
                                 NFC_NDEF_DESFIRE_CC_EF_ID,
                                 14,
                                 1,
                                 &cc_lock);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Update CC file failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

		if (nfc_verbose)
		  printf("Locking the CC file\n");

	  rc = SCardDesfire_ChangeFileSettings(nfc_card_handle(card),
		                                     NFC_NDEF_DESFIRE_CC_EF_ID,
																				 0,
  																			 0xEFFF);
		if (rc != SCARD_S_SUCCESS)
		{
			printf("Change file settings failed\n");
			if (!nfc_silent)
				printf("\t%s\n", SCardDesfire_GetErrorMessage(rc));
			if (p_batch && p_continue)
				goto batch_end;
			goto failed;
		}

    if (!nfc_silent)
		  printf("Card lock done...\n");
	}


batch_end:

  SCardDesfire_DetachLibrary(nfc_card_handle(card));

  if (p_batch)
	{
		if (!nfc_silent)
			printf("Remove the card from the reader...\n");
		if (!nfc_reader_wait_card_absent(p_reader))
		{
			printf("Failed to wait for card removal.\n");
			goto failed;
		}

		goto batch_begin;
	}

  nfc_card_close(card);
  return EXIT_SUCCESS;
  
  /* Error */
failed:
  if (nfc_card_handle(card))
    SCardDesfire_DetachLibrary(nfc_card_handle(card));
  nfc_card_close(card);
	return EXIT_FAILURE;
}



static void usage(void)
{
  printf("\n");
  printf("Prepare the Desfire EV1 card to hold a SmartPoster:\n\n");
	printf("%s --prepare [OPTIONS] , -p [OPTIONS]\n");
	printf("OPTIONS are:\n");
	printf(" --size <Size> , -z <Size>\n");
	printf("\tSize of the NDEF file to be created\n");
	printf("\t(default is to take all the free space in the card)\n");
	printf(" --key <Key> , -k <Key>\n");
	printf("\tCurrent value for root key (default is blank key)\n");
	printf(" --key-type <Type> , -t <Type>\n");
	printf("\nCurrent type for root key (default is DES)\n");
	printf(" --set-key  <Key>\n");
	printf("\tNew value for root key (default is unchanged)\n");
	printf(" --set-key-type <Type>\n");
	printf("\tNew type for root key (default is unchanged)\n");
	printf("\n");
	printf("Lock the Desfire EV1 SmartPoster:\n\n");
	printf("%s --lock [OPTIONS] , -l [OPTIONS]\n");
  printf("\n");
	printf("Erase the Desfire EV1 card:\n\n");
	printf("%s --erase [OPTIONS] , -b [OPTIONS]\n");
	printf("OPTIONS are:\n");
	printf(" --key <Key> , -k <Key>\n");
	printf("\tCurrent value for root key (default is blank key)\n");
	printf(" --key-type <Type> , -t <Type>\n");
	printf("\nCurrent type for root key (default is DES)\n");
	printf(" --set-key  <Key>\n");
	printf("\tNew value for root key (default is unchanged)\n");
	printf(" --set-key-type <Type>\n");
	printf("\tNew type for root key (default is unchanged)\n");
	printf("\n");
	printf("Common options:\n\n");
	printf(" --confirm , -y\n");
	printf("\tDo not ask for confirmation\n");
	printf(" --verbose , -v\n");
	printf("\tGive details on what's going on.\n\n");
	printf(" --quiet , -q\n");
	printf("\tShut up unless fatal errors occur.\n\n");
	printf(" --help , -h\n");
	printf("\tDisplay this help screen, then exits.\n\n");
  printf("PC/SC readers:\n\n");
  printf(" --list-readers , -l\n");
	printf("\tPrint the list of available readers, then exits\n\n");
	printf(" --reader <Reader> , -r <Reader>\n");
	printf("\tWork with the specified Reader. It could be either the\n");
	printf("\tcomplete name of the Reader, or its index in the list.\n\n");
}

static void hello(void)
{
  printf("%s: Format Desfire EV1 cards to hold NFC Forum compliant Tags (SmartPoster)\n", prog_name);
	printf("\n");
  printf("This software is copyright (c) 2000-2011 PRO ACTIVE SAS.\n");
	printf("All rights reserved.\n");
  printf("Visit www.springcard.com/solutions/nfc for information and updates\n");
  printf("\n");
  printf("THIS SOFTWARE IS PROVIDED \"AS IS\" WITH NO EXPRESSED OR IMPLIED WARRANTY.\n");
  printf("You're using it at your own risks.\n");
  printf("\n");
}

static void helpmsg(const char *str)
{
  if (!nfc_silent)
	{
	  if (str != NULL)
		{
      printf(str);
			printf("\n");
		}
    printf("Call %s -h for Help\n", prog_name);
	}
}
