/**h* NFCTools/nfc_pcsc_readers.c
 *
 * NAME
 *   nfc_pcsc_readers.c
 *
 * DESCRIPTION
 *   Reader abstraction layer, on top of PC/SC
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"

#include <winscard.h>
#ifdef WIN32
#pragma comment(lib,"winscard.lib")
#endif

static void nfc_pcsc_cleanup(void);
static void nfc_pcsc_startup(void);

static BOOL nfc_pcsc_started = FALSE;
static int nfc_pcsc_readers_count = 0;
static char **nfc_pcsc_readers_list = NULL;

#ifndef PCSC_E_NO_READERS_AVAILABLE
#define PCSC_E_NO_READERS_AVAILABLE  0x8010002E
#endif

void nfc_list_readers(void)
{
  int i;

  nfc_pcsc_startup();

	switch (nfc_pcsc_readers_count)
	{
	  case 0  : printf("No reader\n");
		          break;
	  case 1  : printf("1 reader:\n");
		          break;
    default : printf("%d readers:\n", nfc_pcsc_readers_count);
		          break;
  }

	for (i=0; i<nfc_pcsc_readers_count; i++)
    printf("\t%3d : %s\n", i + 1, nfc_pcsc_readers_list[i]);

	printf("\n");
}

static void nfc_pcsc_startup(void)
{
  SCARDCONTEXT hContext;
  LONG rc;
	int i;
	char *p;
  char *reader_names = NULL;
	DWORD reader_names_sz = 0;

	if (nfc_pcsc_started)
	  return;

  rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
  if(rc != SCARD_S_SUCCESS)
  {
    printf("SCardEstablishContext : %08lX\n",rc);
	  goto failed;
  }

	rc = SCardListReaders(hContext, NULL, NULL, &reader_names_sz);

	if (rc == (LONG) PCSC_E_NO_READERS_AVAILABLE)
	{
	  /* No reader - nothing to do */
		goto done;
	}
	
	if (rc != SCARD_S_SUCCESS)
	{
	  /* Fatal error */
    printf("SCardListReaders : %08lX\n",rc);
		SCardReleaseContext(hContext);
	  goto failed;
	}
	
	reader_names = nfc_alloc(reader_names_sz);

	rc = SCardListReaders(hContext, NULL, reader_names, &reader_names_sz);

	if (rc != SCARD_S_SUCCESS)
	{
	  /* Fatal error */
    printf("SCardListReaders error : %08lX\n",rc);
		SCardReleaseContext(hContext);
	  goto failed;
	}

  /* Count the readers. */
  p = reader_names;
	for (i=0; *p != '\0'; i++)
	{
		p += strlen(p) + 1;
	}
  nfc_pcsc_readers_count = i;

  nfc_pcsc_readers_list = nfc_alloc(nfc_pcsc_readers_count * sizeof(char *));

  /* Store the names */
	p = reader_names;
	for (i=0; i<nfc_pcsc_readers_count; i++)
	{
    nfc_pcsc_readers_list[i] = nfc_strdup(p);
    p += strlen(p) + 1;
	}

done:
  SCardReleaseContext(hContext);
  nfc_free(reader_names);
	nfc_pcsc_started = TRUE;
	atexit(nfc_pcsc_cleanup);
	return;

failed:
  nfc_free(reader_names);
  exit(EXIT_FAILURE);
}

const char *nfc_reader_find(const char *reader_name)
{
  int i;

  nfc_pcsc_startup();

	if (reader_name == NULL)
	{
    if (nfc_pcsc_readers_count)
		  return nfc_pcsc_readers_list[0];
		return NULL;
	}

	for (i=0; i<nfc_pcsc_readers_count; i++)
	{
#ifdef WIN32
		if (! _stricmp(nfc_pcsc_readers_list[i], reader_name))
			return reader_name;
#else
		if (!strcasecmp(nfc_pcsc_readers_list[i], reader_name))
			return reader_name;
#endif
	}

	i = atoi(reader_name);

	if ((i > 0) && (i <= nfc_pcsc_readers_count))
	  return nfc_pcsc_readers_list[i - 1];

	return NULL;
}

void nfc_pcsc_cleanup(void)
{
  int i;

  if (nfc_pcsc_readers_list != NULL)
	{
    for (i=0; i<nfc_pcsc_readers_count; i++)
		  nfc_free(nfc_pcsc_readers_list[i]);
    nfc_free(nfc_pcsc_readers_list);
	}
}

BOOL nfc_reader_is_card_present(const char *reader_name)
{
  SCARDCONTEXT hContext;
	SCARD_READERSTATE stState;
	LONG rc;

  rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
  if(rc != SCARD_S_SUCCESS)
  {
    printf("SCardEstablishContext : %08lX\n",rc);
	  goto failed;
  }

  stState.szReader = reader_name;
	stState.dwCurrentState = SCARD_STATE_UNAWARE;

	rc = SCardGetStatusChange(hContext, 0, &stState, 1);	
	if (rc != SCARD_S_SUCCESS)
  {
    printf("SCardGetStatusChange : %08lX\n",rc);
    SCardReleaseContext(hContext);
		goto failed;
  }

	SCardReleaseContext(hContext);

	if (stState.dwEventState & SCARD_STATE_PRESENT)
	  return TRUE;
	return FALSE;

failed:
  exit(EXIT_FAILURE);
	return FALSE;
}

BOOL nfc_reader_wait_card_present(const char *reader_name)
{
  SCARDCONTEXT hContext;
	SCARD_READERSTATE stState;
	LONG rc;

  rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
  if(rc != SCARD_S_SUCCESS)
  {
    printf("SCardEstablishContext : %08lX\n",rc);
	  goto failed;
  }

  stState.szReader = reader_name;
	stState.dwCurrentState = SCARD_STATE_UNAWARE;

again:

	rc = SCardGetStatusChange(hContext, INFINITE, &stState, 1);	
	if (rc != SCARD_S_SUCCESS)
  {
    printf("SCardGetStatusChange : %08lX\n",rc);
    SCardReleaseContext(hContext);
		goto failed;
  }


	if (stState.dwEventState & SCARD_STATE_PRESENT)
	{
	  SCardReleaseContext(hContext);
	  return TRUE;
	}

	stState.dwCurrentState = stState.dwEventState;
	goto again;

failed:
  exit(EXIT_FAILURE);
	return FALSE;
}

BOOL nfc_reader_wait_card_absent(const char *reader_name)
{
  SCARDCONTEXT hContext;
	SCARD_READERSTATE stState;
	LONG rc;

  rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
  if(rc != SCARD_S_SUCCESS)
  {
    printf("SCardEstablishContext : %08lX\n",rc);
	  goto failed;
  }

  stState.szReader = reader_name;
	stState.dwCurrentState = SCARD_STATE_UNAWARE;

again:

	rc = SCardGetStatusChange(hContext, INFINITE, &stState, 1);	
	if (rc != SCARD_S_SUCCESS)
  {
    printf("SCardGetStatusChange : %08lX\n",rc);
    SCardReleaseContext(hContext);
		goto failed;
  }


	if (!(stState.dwEventState & SCARD_STATE_PRESENT))
	{
	  SCardReleaseContext(hContext);
	  return TRUE;
	}

	stState.dwCurrentState = stState.dwEventState;
	goto again;

failed:
  exit(EXIT_FAILURE);
	return FALSE;
}


#ifdef WIN32
  #define IOCTL_VENDOR_ESCAPE_SPRINGCARD      SCARD_CTL_CODE(2048)
  #define IOCTL_VENDOR_ESCAPE_MS_CCID         SCARD_CTL_CODE(3500)

	BOOL use_ms_ioctl         = TRUE;
	BOOL use_springcard_ioctl = TRUE;
#endif

#ifdef __linux__
  #include <reader.h>
  #define IOCTL_SMARTCARD_VENDOR_IFD_EXCHANGE SCARD_CTL_CODE(1)
#endif

BOOL nfc_reader_control(const char *reader_name, const BYTE *abCommand, DWORD dwCommandLength, BYTE *abResponse, DWORD *dwResponseLength)
{
  SCARDCONTEXT hContext;
  SCARDHANDLE hCard;
  DWORD dwProtocol;
  LONG rc;
  DWORD MaxLen = *dwResponseLength;
  DWORD CurLen = *dwResponseLength;

	rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
  if(rc != SCARD_S_SUCCESS)
  {
    printf("SCardEstablishContext : %08lX\n",rc);
	  goto failed;
  }

  /* Connect to the card */
  rc = SCardConnect(hContext, reader_name, SCARD_SHARE_DIRECT, 0, &hCard, &dwProtocol);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardConnect(%s) error %08lX\n", reader_name, rc);
		SCardReleaseContext(hContext);
		return FALSE;
	}

#ifdef WIN32
  rc = SCardControl(hCard, IOCTL_VENDOR_ESCAPE_SPRINGCARD, abCommand, dwCommandLength, abResponse, MaxLen, &CurLen);
	if ((rc == ERROR_INVALID_FUNCTION) || (rc == ERROR_NOT_SUPPORTED) || (rc == RPC_X_BAD_STUB_DATA))
	{
		CurLen = *dwResponseLength;
    rc = SCardControl(hCard, IOCTL_VENDOR_ESCAPE_MS_CCID, abCommand, dwCommandLength, abResponse, MaxLen, &CurLen);
	}
#endif

#ifdef __linux__
  rc = SCardControl(hCard, IOCTL_SMARTCARD_VENDOR_IFD_EXCHANGE, abCommand, dwCommandLength, abResponse, MaxLen, &CurLen);
#endif

  SCardDisconnect(hCard, SCARD_LEAVE_CARD);
  SCardReleaseContext(hContext);

  if (rc != SCARD_S_SUCCESS)
  {
    printf("SCardControl error %08lX\n", rc);

#ifdef WIN32
    printf("\n");
    printf("If using MS's CCID driver, make sure you've allowed the 'Vendor Escape'\n");
		printf("in registry. You may use software ms_ccid_escape_enable to do so.\n");
		printf("\n");
#endif
#ifdef __linux__
    printf("\n");
    printf("Make sure you've allowed the 'Vendor Escape' in Info.plist\n");
		printf("Locate the file Info.plist (command: 'locate Info.plist')\n");
		printf("Edit the file (you must be root to do so)\n");
		printf("Locate the line '<ifdDriverOptions>', set bit 1 to 1 to\n");
		printf("allow the 'CCID Exchange' : '<string>0x0001<string>'.\n");
		printf("Restart the PC/SC service after saving file Info.plist .\n");
		printf("\n");
#endif

    goto failed;
  }

  *dwResponseLength = CurLen;
	return TRUE;

failed:
  exit(EXIT_FAILURE);
	return FALSE;
}

BOOL nfc_reader_set_leds(const char *reader_name, BYTE red, BYTE green, BYTE blue)
{
  BYTE c_apdu[256];
  DWORD c_apdu_len = 0;
  BYTE r_apdu[256];
  DWORD r_apdu_len;

  c_apdu[c_apdu_len++] = 0x58;
  c_apdu[c_apdu_len++] = 0x1E;
  c_apdu[c_apdu_len++] = red;
  c_apdu[c_apdu_len++] = green;
  c_apdu[c_apdu_len++] = blue;

  r_apdu_len = sizeof(r_apdu);
  if (!nfc_reader_control(reader_name, c_apdu, c_apdu_len, r_apdu, &r_apdu_len))
    return FALSE;

  return TRUE;
}

BOOL nfc_reader_set_leds_default(const char *reader_name)
{
  BYTE c_apdu[256];
  DWORD c_apdu_len = 0;
  BYTE r_apdu[256];
  DWORD r_apdu_len;

  c_apdu[c_apdu_len++] = 0x58;
  c_apdu[c_apdu_len++] = 0x1E;

  r_apdu_len = sizeof(r_apdu);
  if (!nfc_reader_control(reader_name, c_apdu, c_apdu_len, r_apdu, &r_apdu_len))
    return FALSE;

  return TRUE;
}

BOOL nfc_reader_start_emul(const char *reader_name)
{
  BYTE c_apdu[256];
  DWORD c_apdu_len;
  BYTE r_apdu[256];
  DWORD r_apdu_len;

  c_apdu_len = 0;
  c_apdu[c_apdu_len++] = 0x83;
  c_apdu[c_apdu_len++] = 0x02;
  r_apdu_len = sizeof(r_apdu);
  if (!nfc_reader_control(reader_name, c_apdu, c_apdu_len, r_apdu, &r_apdu_len))
    return FALSE;

  c_apdu_len = 0;
  c_apdu[c_apdu_len++] = 0x83;
  c_apdu[c_apdu_len++] = 0xFD;
  r_apdu_len = sizeof(r_apdu);
  if (!nfc_reader_control(reader_name, c_apdu, c_apdu_len, r_apdu, &r_apdu_len))
    return FALSE;

  return TRUE;
}

BOOL nfc_reader_exit_with_emul(const char *reader_name)
{
  BYTE c_apdu[256];
  DWORD c_apdu_len;
  BYTE r_apdu[256];
  DWORD r_apdu_len;

  c_apdu_len = 0;
  c_apdu[c_apdu_len++] = 0x83;
  c_apdu[c_apdu_len++] = 0x02;
  r_apdu_len = sizeof(r_apdu);
  if (!nfc_reader_control(reader_name, c_apdu, c_apdu_len, r_apdu, &r_apdu_len))
    return FALSE;

  return TRUE;
}

BOOL nfc_reader_stop_emul(const char *reader_name)
{
  BYTE c_apdu[256];
  DWORD c_apdu_len;
  BYTE r_apdu[256];
  DWORD r_apdu_len;

  c_apdu_len = 0;
  c_apdu[c_apdu_len++] = 0x83;
  c_apdu[c_apdu_len++] = 0xFE;
  r_apdu_len = sizeof(r_apdu);
  if (!nfc_reader_control(reader_name, c_apdu, c_apdu_len, r_apdu, &r_apdu_len))
    return FALSE;

  c_apdu_len = 0;
  c_apdu[c_apdu_len++] = 0x83;
  c_apdu[c_apdu_len++] = 0xFF;
  r_apdu_len = sizeof(r_apdu);
  if (!nfc_reader_control(reader_name, c_apdu, c_apdu_len, r_apdu, &r_apdu_len))
    return FALSE;

  return TRUE;
}
