/**h* NFCTools/nfc_type_2.c
 *
 * NAME
 *   nfc_type_1.c
 *
 * DESCRIPTION
 *   Support of NFC Type 2 Tag
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"
#include "nfc_card_i.h"

BOOL nfc_card_ultralight_recognize(nfc_card_t card)
{
	unsigned char buffer[16];
	unsigned char buffer_t[16];
	int i, page, pages;

  if (nfc_verbose)
		printf("\tthis is a chip in the Mifare UltraLight family\n");

	card->type = CARD_NXP_MIFARE_UL;
	card->capacity = 48;
	pages = 16;

  /* Recognize the chip */
  if (!nfc_card_read_memory(card,  0, buffer, 16))
	{
		if (nfc_verbose)
			printf("\tfailed to read pages 0-3\n");
		return FALSE;
	}
  if (!nfc_card_read_memory(card, 16, buffer_t, 16))
	{
		if (nfc_verbose)
			printf("\tpages 16-19 not readable, this must be a Mifare UltraLight\n");
	} else
	if (!memcmp(buffer, buffer_t, 16))
	{
		if (nfc_verbose)
			printf("\tpages 0-3 and 16-19 are equals, this must be a Mifare UltraLight\n");
	} else
	{
		if (nfc_verbose)
			printf("\tpages 0-3 and 16-19 are different, this must be a Mifare UltraLight C\n");
		card->type = CARD_NXP_MIFARE_UL_C;
		card->capacity = 134; //144-10 bytes for lock and memory control (5 bytes each)
		pages = 48;
	}

	card->writable = TRUE;

	/* Analyze the CC - it must either be empty, or a valid CC */
	if ((buffer[12] == NFC_FORUM_MAGIC_NUMBER) && (buffer[13] != 0x00) && (buffer[14] != 0x00))
	{
		card->prepared = TRUE;
    //card->capacity = 8 * buffer[14];

		if (nfc_verbose)
			printf("\tCC are valid, capacity=%dB\n", card->capacity);

		if ((buffer[15] & 0x0F) == 0x0F)
		{
			card->writable = FALSE;
			if (nfc_verbose)
        printf("\tCC says the card is not writable\n");
		}
	} else
	if ((buffer[12] == 0x00) || (buffer[13] == 0x00) || (buffer[14] == 0x00) || (buffer[15] == 0x00))
	{
		if (nfc_verbose)
      printf("\tCC are blank\n");
	} else
	{
		card->writable = FALSE;
		if (nfc_verbose)
      printf("\tCC are invalid\n");
	}

	/* Check the locks */
	if (card->writable)
	{
		if (buffer[10] || buffer[11])
		{
		  card->writable = FALSE;
			if (nfc_verbose)
				printf("\tlock bits of page 3 are set\n");
		}
	}
	if (card->writable && (card->type == CARD_NXP_MIFARE_UL_C))
	{
		if (!nfc_card_read_memory(card, 40, buffer, 16))
		{
			if (nfc_verbose)
				printf("\tread page 40 failed\n");
		} else
		if (buffer[0] || buffer[1])
		{
		  card->writable = FALSE;
			if (nfc_verbose)
				printf("\tlock bits of page 40 are set\n");
		}
	}

	/* Is card empty ? */
	card->empty = TRUE;
	for (page=4; page<pages; page+=4)
	{
		if (!nfc_card_read_memory(card, (unsigned short) page, buffer, 16))
		{
			if (nfc_verbose)
				printf("\tread page 40 failed\n");
			break;
		}

		for (i=0; i<16; i++)
		{
			if (buffer[i] != 0x00)
			{
				if (nfc_verbose)
				  printf("\tcard is not empty\n");
				card->empty = FALSE;
				break;
      }
		}

		printf("\n");

		if (!card->empty)
			break;
	}

  /* Is MyFare UltraLight C correctly formatted ? */
  if (card->type == CARD_NXP_MIFARE_UL_C)
  {
    if(!nfc_card_read_memory(card,  4, buffer, 16))
    {
			if (nfc_verbose)
				printf("\tRead page 5 on UltraLight C failed\n");
    }
          //  card->prepared = FALSE;
    if ( (buffer[0] != 0x01) || (buffer[1] != 0x03) || (buffer[5] != 0x02) || (buffer[6] != 0x03) || (buffer[10] != 0x03) )
      {
        if (nfc_verbose)
          printf("\tUltraLight C card not correctly formatted\n");
         // printf("buffer0: %02x - buffer1: %02x - buffer5: %02x - buffer6: %02x - buffer10: %02x\n", (int)buffer[0],(int)buffer[1],(int)buffer[5],(int)buffer[6],(int)buffer[10]);
        card->prepared = FALSE;
      }
  }


	return TRUE;
}

BOOL nfc_card_ultralight_prepare(nfc_card_t card)
{
  //BYTE buffer[20];

  if ((card == NULL) || ((card->type != CARD_NXP_MIFARE_UL) && (card->type != CARD_NXP_MIFARE_UL_C)))
    return FALSE;

  if (card->type == CARD_NXP_MIFARE_UL)
  {
    BYTE buffer1[4];
    buffer1[0] = NFC_FORUM_MAGIC_NUMBER;
	  buffer1[1] = NFC_FORUM_VERSION_NUMBER;
	  buffer1[2] = card->capacity / 8;
	  buffer1[3] = 0x00; 

   	if (nfc_verbose)
		  printf("\twriting the CC on page 3\n");

	  if (!nfc_card_write_memory(card, 3, buffer1, 4))
	  {
	    if (!nfc_silent)
		    printf("Failed to write the CC into the UltraLight card\n");
		  return FALSE;
	  }
    
    // TODO : inserer un NDEF vide
    buffer1[0] = 0x03;
	  buffer1[1] = 0x00;
	  buffer1[2] = 0xFE; //Terminator
	  buffer1[3] = 0x00; 

    if (nfc_verbose)
		  printf("\twriting the first empty NDEF and Terminator on page 4\n");

	  if (!nfc_card_write_memory(card, 4, buffer1, 4))
	  {
	    if (!nfc_silent)
		    printf("Failed to write the first empty NDEF and Terminator into the UltraLight card\n");
		  return FALSE;
	  }

  } else
  if (card->type == CARD_NXP_MIFARE_UL_C)
  {
    BYTE buffer2[4];

    //printf("on est ici\n");
    buffer2[0] = NFC_FORUM_MAGIC_NUMBER;
	  buffer2[1] = NFC_FORUM_VERSION_NUMBER;
	  buffer2[2] = 0x12;
	  buffer2[3] = 0x00;
    if (nfc_verbose)
	   printf("\twriting page 3\n");

    if (!nfc_card_write_memory(card, 3, buffer2, 4))
	  {
       if (!nfc_silent)
		      printf("Failed to write the CC into the UltraLight-C card\n");
		    return FALSE;
    }

    buffer2[0] = 0x01;
    buffer2[1] = 0x03;
    buffer2[2] = 0xA0;
    buffer2[3] = 0x0C;
    if (nfc_verbose)
	   printf("\twriting page 4\n");

    if (!nfc_card_write_memory(card, 4, buffer2, 4))
	  {
       if (!nfc_silent)
		      printf("Failed to write the lock control into the UltraLight-C card\n");
		    return FALSE;
    }

    buffer2[0] = 0x24;
    buffer2[1] = 0x02;
    buffer2[2] = 0x03;
    buffer2[3] = 0xA2;
    if (nfc_verbose)
	   printf("\twriting page 5\n");

    if (!nfc_card_write_memory(card, 5, buffer2, 4))
	  {
       if (!nfc_silent)
		      printf("Failed to write the memory control into the UltraLight-C card\n");
		    return FALSE;
    }

    buffer2[0] = 0x1E;
    buffer2[1] = 0x04;
    buffer2[2] = 0x03;
    buffer2[3] = 0x00;
    if (nfc_verbose)
	   printf("\twriting page 6\n");

    if (!nfc_card_write_memory(card, 6, buffer2, 4))
	  {
       if (!nfc_silent)
		      printf("Failed to write the first empty NDEF TLV and Terminator into the UltraLight-C card\n");
		    return FALSE;
    }

    buffer2[0] = 0xFE;
    buffer2[1] = 0x00;
    buffer2[2] = 0x00;
    buffer2[3] = 0x00;
    if (nfc_verbose)
	   printf("\twriting page 7\n");

    if (!nfc_card_write_memory(card, 7, buffer2, 4))
	  {
       if (!nfc_silent)
		      printf("Failed to write the Terminator TLV into the UltraLight-C card\n");
		    return FALSE;
    }
  //  nfc_card_read_memory(card,  6, buffer2, 4);
  //  printf("UltralLight C : bit0: %02x - bit1: %02x - bit2: %02x - bit3: %02x\n", (int)buffer2[0], (int)buffer2[1], (int)buffer2[2], (int)buffer2[3]);

  }

	return TRUE;
}

BOOL nfc_card_ultralight_lock(nfc_card_t card)
{
  BYTE buffer[4];

  if ((card == NULL) || ((card->type != CARD_NXP_MIFARE_UL) && (card->type != CARD_NXP_MIFARE_UL_C)))
    return FALSE;

  if (card->type == CARD_NXP_MIFARE_UL_C)
	{
    buffer[0] = 0xFF;
		buffer[1] = 0xFF;
		buffer[2] = 0x00;
		buffer[3] = 0x00;

		if (!nfc_card_write_memory(card, 40, buffer, 4))
		{
      /* Write could fail as we are ORing a WORM area */
      if (nfc_card_read_memory(card, 40, buffer, 4) && (buffer[0] == 0xFF) && (buffer[1] == 0xFF))
      {
        /* Consider there's no error... */
        if (nfc_verbose)
          printf("(write did succeed anyway)\n");

      } else
      {
			  if (!nfc_silent)
				  printf("Failed to write page 40 of the UltraLight C card\n");
			  return FALSE;
      }
		}
	}

  buffer[0] = 0x00;
	buffer[1] = 0x00;
	buffer[2] = 0xFF;
	buffer[3] = 0xFF;

	if (!nfc_card_write_memory(card, 2, buffer, 4))
	{
    /* Write could fail as we are ORing a WORM area */
    if (nfc_card_read_memory(card, 2, buffer, 4) && (buffer[2] == 0xFF) && (buffer[3] == 0xFF))
    {
      /* Consider there's no error... */
      if (nfc_verbose)
        printf("(write did succeed anyway)\n");

    } else
    {
		  if (!nfc_silent)
			  printf("Failed to write page 2 of the UltraLight%s card\n", (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
		  return FALSE;
    }
	}

  buffer[0] = 0x00;
	buffer[1] = 0x00;
	buffer[2] = 0x00;
	buffer[3] = 0x0F;

	if (!nfc_card_write_memory(card, 3, buffer, 4))
	{
    /* Write could fail as we are ORing a WORM area */
    if (nfc_card_read_memory(card, 2, buffer, 4) && ((buffer[3] & 0x0F) == 0x0F))
    {
      /* Consider there's no error... */
      if (nfc_verbose)
        printf("(write did succeed anyway)\n");

    } else
    {
		  if (!nfc_silent)
			  printf("Failed to write page 3 of the UltraLight%s card\n", (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
		  return FALSE;
    }
	}

  return TRUE;
}

BOOL nfc_card_ultralight_write_content(nfc_card_t card, unsigned char *content, size_t content_length)
{
 
  
  // TODO : relire de la page 4 jusqu'a trouver le TLV NDEF vide  
  // TODO : ecrire le content en ECRASANT le TLV NDEF
  
  BYTE buffer[4];
  BYTE buf[2];
	int page, last_page;
	size_t offset = 0;

  if (card == NULL)
    return FALSE;
    
  if ((card->type != CARD_NXP_MIFARE_UL) && (card->type != CARD_NXP_MIFARE_UL_C))
    return FALSE;


  if (card->type == CARD_NXP_MIFARE_UL)
  {
    nfc_card_read_memory(card,  4, buffer, 4);
    // printf("UltraLight : bit0: %X - bit1: %X - bit2: %X - bit3: %X\n", buffer[0], buffer[1], buffer[2], buffer[3]);
    //block4, bit 0 = 0x03 => static

    last_page = 4 + card->capacity / 4;

    for (page=4; page<last_page; page++)
    {
      if (offset > content_length)
        break;

      memcpy(buffer, &content[offset], 4);
      offset += 4;

      if (!nfc_card_write_memory(card, (unsigned short) page, buffer, 4))
      {
        if (!nfc_silent)
          printf("Failed to write page %d of the UltraLight%s card\n", page, (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
        return FALSE;
      }
    }
   return TRUE;    
  }

  if (card->type == CARD_NXP_MIFARE_UL_C)
  {
    nfc_card_read_memory(card,  6, buffer, 4);
    //printf("UltralLight C : bit0: %02x - bit1: %02x - bit2: %02x - bit3: %02x\n", (int)buffer[0], (int)buffer[1], (int)buffer[2], (int)buffer[3]);

    memcpy(buf, &content[0], 2);
    last_page = 40;

    /* write page 6 */
    nfc_card_read_memory(card,  6, buffer, 4); // to get bytes 0 and 1
    
    
    buffer[2] = buf[0];
    buffer[3] = buf[1];
    offset = 2;
    if (!nfc_card_write_memory(card, 6, buffer, 4))
      {
        if (!nfc_silent)
          printf("Failed to write page 6 of the UltraLight%s card\n", (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
        return FALSE;
      }
  


    /* from page 7 until the end */
    for (page=7; page<last_page; page++)
    {
      if (offset > content_length)
        break;

      memcpy(buffer, &content[offset], 4);
      offset += 4;
 

      if (!nfc_card_write_memory(card, (unsigned short) page, buffer, 4))
      {
        if (!nfc_silent)
          printf("Failed to write page %d of the UltraLight%s card\n", page, (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
        return FALSE;
      }
    }
    return TRUE;
   
  }



  return TRUE;
}

static const char *sorry = "Sorry, handling of Mifare Classic cards is not yet implemented\n";

BOOL nfc_card_mifare_recognize(nfc_card_t card)
{
  (void) (card);

  printf(sorry);
  return FALSE;
}

BOOL nfc_card_mifare_prepare(nfc_card_t card)
{
  (void) (card);

  printf(sorry);
  return FALSE;
}

BOOL nfc_card_mifare_lock(nfc_card_t card)
{
  (void) (card);

  printf(sorry);
  return FALSE;
}

BOOL nfc_card_mifare_write_content(nfc_card_t card, unsigned char *content, size_t content_length)
{
  (void) (card);
	(void) (content);
	(void) (content_length);

  printf(sorry);
  return FALSE;
}
