/**h* NFCTools/nfc_utils.c
 *
 * NAME
 *   nfc_utils.c
 *
 * DESCRIPTION
 *   Memory and file-system related functions
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"

#include <sys/stat.h>
#include <errno.h>

void *nfc_alloc(size_t size)
{
  void *result;
	result = malloc(size);
	if (result == NULL)
	{
	  fprintf(stderr, "Out of memory (can't allocate %luB)\n", (unsigned long) size);
    exit(EXIT_FAILURE);
	} else
	{
    memset(result, 0, size);
	}
	return result;
}

char *nfc_strdup_blob(unsigned char *blob, size_t size)
{
	char *result = NULL;
  if (blob != NULL)
	{
    result = nfc_alloc(size + 1);
    memcpy(result, blob, size);
  }
	return result;
}

char *nfc_strdup(char *str)
{
  size_t size;
	char *result = NULL;
  if (str != NULL)
	{
		size = strlen(str);
		result = nfc_alloc(size + 1);
		strcpy(result, str);
	}
	return result;
}

char *nfc_strdup_length(char *str, size_t *size)
{
  char *result = nfc_strdup(str);
	if (size != NULL)
	{
	  if (result != NULL)
		  *size = strlen(result);
		else
      *size = 0;
	}
	return result;
}

void nfc_free(void *ptr)
{
  if (ptr != NULL)
	  free(ptr);
}

char *nfc_file_name_alloc(char *file_name, char *default_extension)
{
  size_t l;
	char *r;

  if (file_name == NULL)
	  return NULL;

	if ((default_extension == NULL) || (strchr(file_name, '.') != NULL))
	  return nfc_strdup(file_name);

	l = strlen(file_name) + 1 + strlen(default_extension) + 1;
  r = nfc_alloc(l);

	sprintf(r, "%s.%s", file_name, default_extension);
  return r;
}

char *nfc_file_type_alloc(char *file_name)
{  
	FILE *fp;
	char line[128];
	char *extension;
	
	if (file_name == NULL)
	  return NULL;

	extension = strrchr(file_name, '.');
	if (extension == NULL)
	  return NULL;
	extension++;
	if (!strlen(extension))
	  return NULL;
	
	fp = fopen("mime.types", "rt");
	if (fp == NULL)
	  return NULL;

  while (fgets(line, sizeof(line), fp) != NULL)
	{
	  char *p_extension = strstr(line, extension);

	  if (p_extension != NULL)
		{
		  p_extension = strtok(p_extension, " \t");
			if (!strcmp(p_extension, extension))
			{
			  char *p_mime_type = strtok(line, " \t");
				fclose(fp);
				return nfc_strdup(p_mime_type);
			}
		}
	}

	fclose(fp);
	return NULL;
}

BOOL nfc_file_exists(char *file_name)
{
  struct stat fs;

  if (stat(file_name, &fs) < 0)
	{
    if(errno != ENOENT)
		{
      if (!nfc_silent)
			  perror("file_exists(stat)");
    }
		return FALSE;
  }
  return TRUE;
}

size_t nfc_file_size(char *file_name)
{
  struct stat fs;

  if (stat(file_name, &fs) < 0)
	{
    if(errno != ENOENT)
		{
      if (!nfc_silent)
			  perror("file_size(stat)");
    }
		return 0;
  }
  return fs.st_size;
}

BOOL nfc_read_file(unsigned char *blob, size_t size, const char *file_name)
{
  FILE *fp;
  size_t got, offset = 0;

  fp = fopen(file_name, "rb");
  if (fp == NULL)
    return FALSE;

  while (offset < size)
  {
    got = fread(&blob[offset], 1, size - offset, fp);
    if (got <= 0)
    {
      if (!nfc_silent)
			  printf("read_file: read failed (%d)\n", errno);

      fclose(fp);
      return FALSE;
    }
    offset += got;
  }

  fclose(fp);
  return TRUE;
}

BOOL nfc_write_file(unsigned char *blob, size_t size, const char *file_name)
{
  FILE *fp;
  size_t done, offset = 0;

  fp = fopen(file_name, "wb");
  if (fp == NULL)
    return FALSE;

  while (offset < size)
  {
    done = fwrite(&blob[offset], 1, size - offset, fp);
    if (done <= 0)
    {
      if (!nfc_silent)
			  printf("write_file: write failed (%d)\n", errno);

      fclose(fp);
      return FALSE;
    }
    offset += done;
  }

  fclose(fp);
  return TRUE;
}

BOOL nfc_is_hex(char c)
{
  if ((c >= '0') && (c <= '9')) return TRUE;
	if ((c >= 'A') && (c <= 'F')) return TRUE;
	if ((c >= 'a') && (c <= 'f')) return TRUE;
	return FALSE;
}

unsigned char nfc_hex_to_byte(char c)
{
  if ((c >= '0') && (c <= '9')) return c - '0';
	if ((c >= 'A') && (c <= 'F')) return c - 'A' + 10;
	if ((c >= 'a') && (c <= 'f')) return c - 'a' + 10;
	return 0xFF;
}

BOOL nfc_hex_to_bytes(unsigned char *buffer, size_t max_size, size_t *got_size, const char *str, BOOL strict)
{
  size_t i_src = 0, i_dst = 0;

  if ((buffer == NULL) || (str == NULL) || ((got_size == NULL) && !strict))
	  return FALSE;

  while (str[i_src] != '\0')
	{
	  if (nfc_is_hex(str[i_src]))
		{
		  unsigned char b = nfc_hex_to_byte(str[i_src]);

			if (i_dst % 2)
			{
        buffer[i_dst / 2] |= b;
			} else
			{
        buffer[i_dst / 2] = b << 4;
			}
			i_dst++;
			if (i_dst >= max_size)
			  break;
		}
		i_src++;
	}

	if (got_size != NULL)
	  *got_size = i_dst / 2;

	if (strict)
	{
    if ((i_dst % 2) || ((i_dst / 2) < max_size))
		  return FALSE;
	}

	return TRUE;
}
