/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2011 PRO ACTIVE SAS - www.springcard.com


  QuickBatch
  ----------
  
  QuickBatch is a reference source code that shows how to overcome
  PC/SC's bottleneck to fasten-up the process of encoding multiple
  cards in a batch.

  The program consist of two parts:
  - do the processing using only 'standard' PC/SC features
  - do the same processing using SpringCard specific features
  After both parts a summary is displayed to check the illustrate
  the difference in performance.

  For the purpose of the demo, the card processing is a 'dummy' loop
  where a single command is repeated N times :
  - for Memory cards (including Mifare Classic) : read of address 0
  - for T=CL cards : 'select master file' command
  Of course you should replace this part by your own code.

  To speed up the process, the type of card is not determined by the
  software, but has to be defined on the command line. This is
  consistent with the idea of processing multiple cards in a batch,
  under the control of the application, instead of waiting for a
  random card.

  QuickBatch is primarily designed with CrazyWriter in mind (because
  this hardware is suitable to be embedded in card personalization
  machines) but will work as is with other SpringCard readers.

*/

#include "quickbatch.h"

BOOL be_silent  = FALSE;
BOOL be_verbose = TRUE;

typedef struct
{
  unsigned long n_seen;
  unsigned long n_success;
  unsigned long n_failure;
  unsigned long t_total_ms;
  unsigned long t_process_ms;
  unsigned long t_lost_ms;

} PROCESS_BENCHMARK_ST;


BOOL process_loop(char *reader_name, BOOL specific, unsigned long n_cards, unsigned long n_apdus, PROCESS_BENCHMARK_ST *benchmark);
BOOL process_core(smartcard_t smartcard, unsigned long n_apdus);

int main(int argc, char **argv)
{
  char *reader_name;
  PROCESS_BENCHMARK_ST benchmark_standard;
  PROCESS_BENCHMARK_ST benchmark_specific;
  unsigned int N_CARDS = 10;
  unsigned int N_APDUS = 200;
  smartcard_t smartcard;

  reader_name = pcsc_reader_find("1");

  if (reader_name == NULL)
  {
    printf("No such reader\n");
    return EXIT_FAILURE;
  }

  printf("Reader = '%s'\n", reader_name);

  if (pcsc_is_card_present(reader_name))
  {
    printf("Please remove the card that is currently in the reader...\n");
    if (!pcsc_wait_card_absent(reader_name))
    {
      printf("ERROR waiting for card removal\n");
      return FALSE;
    }
  }

  /* To speed up the process, first step is to have the reader looking up    */
  /* only for the type of cards we want to work with (no need to poll for    */
  /* ISO 14443-B if you all your cards are ISO 14443-A, for instance). This  */
  /* is done through SCardControl, so we need a direct access to the reader. */

  smartcard = pcsc_connect_direct(reader_name);
  if (smartcard == NULL)
  {
    printf("Failed to connect to the reader\n");
    return EXIT_FAILURE;
  }

  if (!control_select_card_types(smartcard, TRUE, FALSE))
  {
    printf("Failed to configure reader's polling\n");
    return EXIT_FAILURE;
  }

  pcsc_disconnect(smartcard);

  printf("Ready...\n");

  if (!process_loop(reader_name, FALSE, N_CARDS, N_APDUS, &benchmark_standard))
  {
    printf("Standard mode test failed\n");
    return EXIT_FAILURE;
  }

  if (!process_loop(reader_name, TRUE, N_CARDS, N_APDUS, &benchmark_specific))
  {
    printf("Specific mode test failed\n");
    return EXIT_FAILURE;
  }

  printf("Performance summary                  STANDARD             SPECIFIC\n");
  printf("                                   total   /card      total   /card\n");

  printf("- Cards seen                  : %8lu           %8lu\n", benchmark_standard.n_seen,
                                                                  benchmark_specific.n_success);
  printf("- Cards processed OK          : %8lu           %8lu\n", benchmark_standard.n_success,
                                                                  benchmark_specific.n_success);
  printf("- Cards processed KO          : %8lu           %8lu\n", benchmark_standard.n_failure,
                                                                  benchmark_specific.n_failure);
  printf("- Total processing time       : %8lums %5lums %8lums %5lums\n", benchmark_standard.t_process_ms,
                                                                          benchmark_standard.t_process_ms / N_CARDS,
                                                                          benchmark_specific.t_process_ms,
                                                                          benchmark_specific.t_process_ms / N_CARDS);
  printf("- Total duration of the test  : %8lums         %8lums\n", benchmark_standard.t_total_ms,
                                                                    benchmark_specific.t_total_ms);
  printf("- Time lost in card movements : %8lums %5lums %8lums %5lums\n", benchmark_standard.t_lost_ms,
                                                                          benchmark_standard.t_lost_ms / N_CARDS,
                                                                          benchmark_specific.t_lost_ms,
                                                                          benchmark_specific.t_lost_ms / N_CARDS);
  printf("\n");

  printf("\nDone!\n");
  return EXIT_SUCCESS;
}

BOOL process_loop(char *reader_name, BOOL specific, unsigned long n_cards, unsigned long n_apdus, PROCESS_BENCHMARK_ST *benchmark)
{
  smartcard_t smartcard = NULL;
  BYTE previous_uid[16], previous_uidlen = 0;
  BYTE current_uid[16], current_uidlen;

  LARGE_INTEGER t_process, t_total, t0, t1, f;


  t_total.QuadPart   = 0;
  t_process.QuadPart = 0;
  QueryPerformanceFrequency(&f);
  f.QuadPart /= 1000;

  memset(benchmark, 0, sizeof(PROCESS_BENCHMARK_ST));

  if (!specific)
  {
    /* Only standard PC/SC features */
    /* --------------------------- */

    for (benchmark->n_success=0; benchmark->n_success<n_cards; )
    {
      /* Wait for card arrival - standard */
      /* -------------------------------- */

      printf("[S] Waiting for card arrival (%u/%u).........\n", benchmark->n_success+1, n_cards);

      /* Wait for card arrival using SCardGetStatusChange */
      if (!pcsc_wait_card_present(reader_name))
      {
        printf("[S] ERROR waiting for card arrival\n");
        return FALSE;
      }

      if (benchmark->n_seen == 0)
      {
        /* Initialize the total time here */
        QueryPerformanceCounter(&t_total);
      }

      benchmark->n_seen++;

      /* Initialize the processing time here */
      QueryPerformanceCounter(&t0);

      /* SCardConnect, standard mode */
      smartcard = pcsc_connect(reader_name);
      if (smartcard == NULL)
      {
        printf("[S] Failed to connect to the card\n");
        continue;
      }

      /* A card is present and connected */
      /* -------------------------------- */

      /* Let's make sure this is a new card by reading the UID */
      current_uidlen = sizeof(current_uid);
      if (!smartcard_get_uid(smartcard, current_uid, &current_uidlen))
      {
        printf("[S] Failed to get the UID of the card\n");
      } else
      {
        if ((current_uidlen == previous_uidlen) && !memcmp(current_uid, previous_uid, current_uidlen))
        {
          /* Same UID a earlier - We don't want to process the same card again */
          printf("[S] Same card as earlier, ignoring...\n");
        } else
        {
          printf("[S] Processing the card...\n");

          if (process_core(smartcard, n_apdus))
          {
            benchmark->n_success++;
            printf("[S] OK, remove the card\n");

            /* Remember the UID - we don't want to process the same card again */
            memcpy(previous_uid, current_uid, current_uidlen);
            previous_uidlen = current_uidlen;

          } else
          {
            /* Failure - don't remember the UID. We may remove the card and insert it again */
            benchmark->n_failure++;
            printf("[S] FAILED, remove the card\n");
          }

          /* How long did it took ? */
          QueryPerformanceCounter(&t1);
          t_process.QuadPart += t1.QuadPart - t0.QuadPart;
        }
      }

      /* Wait for card removal - standard */
      /* -------------------------------- */

      /* SCardDisconnect */
      if (!pcsc_disconnect(smartcard))
      {
        printf("[S] ERROR waiting in disconnect\n");
        return FALSE;
      }

      /* Wait for card removal using SCardGetStatusChange */
      if (!pcsc_wait_card_absent(reader_name))
      {
        printf("[S] ERROR waiting for card removal\n");
        return FALSE;
      }
    }

    /* Done with the loop */
    QueryPerformanceCounter(&t1);

  } else
  {
    /* SpringCard specific features */
    /* ---------------------------- */

    for (benchmark->n_success=0; benchmark->n_success<n_cards; )
    {
      /* Wait for card arrival - not standard */
      /* ------------------------------------ */

      printf("[A] Waiting for card arrival (%u/%u).........\n", benchmark->n_success+1, n_cards);

      if (smartcard == NULL)
      {
        /* On the first time, we wait 'classicaly' for card arrival using SCardGetStatusChange */
        if (!pcsc_wait_card_present(reader_name))
        {
          printf("[A] ERROR waiting for card arrival (first)\n");
          return FALSE;
        }

        if (benchmark->n_seen == 0)
        {
          /* Initialize the total time here */
          QueryPerformanceCounter(&t_total);
        }

        benchmark->n_seen++;


        /* SCardConnect, same as in standard mode, but the card will never be reported as removed */
        smartcard = pcsc_connect(reader_name);
        if (smartcard == NULL)
        {
          printf("[A] Failed to connect to the card\n");
          continue;
        }
        
        /* Stop the card tracking - we'll monitor the card manually */
        if (!springcard_suspend_tracking(smartcard))
        {
          printf("[A] Failed to suspend the tracking\n");
          return FALSE;
        }

      } else
      {
        /* We are already connnected to a smartcard (the previous one actually), we just   */
        /* have to wait until the next one arrives. This is done through a pseudo-APDU (in */
        /* a loop) and not through SCardGetStatusChange as card tracking has been disabled */

        if (!springcard_wait_card_present(smartcard))
        {
          printf("[A] ERROR waiting for card arrival (first)\n");
          return FALSE;
        }

        benchmark->n_seen++;
      }

      /* Initialize the processing time here */
      QueryPerformanceCounter(&t0);

      /* A card is present and connected - We are back to standard */
      /* ---------------------------------------------------------- */

      /* Let's make sure this is a new card by reading the UID */
      current_uidlen = sizeof(current_uid);
      if (!smartcard_get_uid(smartcard, current_uid, &current_uidlen))
      {
        printf("[A] Failed to get the UID of the card\n");
      } else
      {
        if ((current_uidlen == previous_uidlen) && !memcmp(current_uid, previous_uid, current_uidlen))
        {
          /* Same UID a earlier - We don't want to process the same card again */
          printf("[A] Same card as earlier, ignoring...\n");
        } else
        {
          printf("[A] Processing the card...\n");

          if (process_core(smartcard, n_apdus))
          {
            benchmark->n_success++;
            printf("[A] OK, remove the card\n");

            /* Remember the UID - we don't want to process the same card again */
            memcpy(previous_uid, current_uid, current_uidlen);
            previous_uidlen = current_uidlen;

          } else
          {
            /* Failure - don't remember the UID. We may remove the card and insert it again */
            benchmark->n_failure++;
            printf("[A] FAILED, remove the card\n");
          }

          /* How long did it took ? */
          QueryPerformanceCounter(&t1);
          t_process.QuadPart += t1.QuadPart - t0.QuadPart;
        }
      }

      /* Wait for card removal - not standard */
      /* ------------------------------------ */

      /* Wait for card removal using pseudo-APDUs in a loop */
      if (!springcard_wait_card_absent(smartcard))
      {
        printf("[S] ERROR waiting for card removal\n");
        return FALSE;
      }
    }

    /* Done with the loop */
    QueryPerformanceCounter(&t1);

    /* Restore reader's standard behaviour */
    if (!springcard_resume_tracking(smartcard))
    {
      printf("[A] Failed to resume the tracking\n");
      return FALSE;
    }

    /* Close the connection to the reader */
    if (!pcsc_disconnect(smartcard))
    {
      printf("[A] ERROR waiting in disconnect\n");
      return FALSE;
    }
  }

  t_total.QuadPart = t1.QuadPart - t_total.QuadPart;

  benchmark->t_process_ms = (unsigned long) (t_process.QuadPart / f.QuadPart);
  benchmark->t_total_ms   = (unsigned long) (t_total.QuadPart / f.QuadPart);
  benchmark->t_lost_ms    = (unsigned long) ((t_total.QuadPart - t_process.QuadPart) / f.QuadPart);

  return TRUE;
}

BOOL process_core(smartcard_t smartcard, unsigned long n_apdus)
{
  BYTE buffer[256];
  unsigned int i, retry = 3;
  BOOL r;

  for (i=0; i<n_apdus; i++)
  {
    if (!smartcard_read_memory(smartcard, 0, buffer, 0))
    {
      if (!be_silent)
        printf("Read address 0 failed, iteration %u/%u\n", i+1, n_apdus);
      r = FALSE;

      if (retry)
      {
        /* Try again */
        i--;
        retry--;
      } else
      {
        return FALSE;
      }
    }
  }

  return TRUE;

}
