#include "quickbatch.h"

#ifdef WIN32
#pragma comment(lib,"winscard.lib")
#endif

#define IOCTL_VENDOR_ESCAPE_SPRINGCARD      SCARD_CTL_CODE(2048)

static void pcsc_cleanup(void);
static void pcsc_startup(void);

static BOOL pcsc_started = FALSE;
static int pcsc_readers_count = 0;
static char **pcsc_readers_list = NULL;
static SCARDCONTEXT scardcontext;

#ifndef PCSC_E_NO_READERS_AVAILABLE
#define PCSC_E_NO_READERS_AVAILABLE  0x8010002E
#endif

void list_pcsc_readers(void)
{
  int i;

  pcsc_startup();

	switch (pcsc_readers_count)
	{
	  case 0  : printf("No reader\n");
		          break;
	  case 1  : printf("1 reader:\n");
		          break;
    default : printf("%d readers:\n", pcsc_readers_count);
		          break;
  }

	for (i=0; i<pcsc_readers_count; i++)
    printf("\t%3d : %s\n", i + 1, pcsc_readers_list[i]);

	printf("\n");
}

static void pcsc_startup(void)
{
  LONG rc;
	int i;
	char *p;
  char *reader_names = NULL;
	DWORD reader_names_sz = 0;

	if (pcsc_started)
	  return;

  rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &scardcontext);
  if(rc != SCARD_S_SUCCESS)
  {
    printf("SCardEstablishContext : %08lX\n",rc);
	  goto failed;
  }

	rc = SCardListReaders(scardcontext, NULL, NULL, &reader_names_sz);

	if (rc == (LONG) PCSC_E_NO_READERS_AVAILABLE)
	{
	  /* No reader - nothing to do */
		goto done;
	}
	
	if (rc != SCARD_S_SUCCESS)
	{
	  /* Fatal error */
    printf("SCardListReaders : %08lX\n",rc);
		SCardReleaseContext(scardcontext);
	  goto failed;
	}
	
	reader_names = malloc(reader_names_sz);
	if (reader_names == NULL)
	{
	  perror("malloc");
    exit(EXIT_FAILURE);
	}

	rc = SCardListReaders(scardcontext, NULL, reader_names, &reader_names_sz);

	if (rc != SCARD_S_SUCCESS)
	{
	  /* Fatal error */
    printf("SCardListReaders error : %08lX\n",rc);
		SCardReleaseContext(scardcontext);
	  goto failed;
	}

  /* Count the readers. */
  p = reader_names;
	for (i=0; *p != '\0'; i++)
	{
		p += strlen(p) + 1;
	}
  pcsc_readers_count = i;

  pcsc_readers_list = malloc(pcsc_readers_count * sizeof(char *));
	if (pcsc_readers_list == NULL)
	{
	  perror("malloc");
    exit(EXIT_FAILURE);
	}

  /* Store the names */
	p = reader_names;
	for (i=0; i<pcsc_readers_count; i++)
	{
    pcsc_readers_list[i] = strdup(p);
    p += strlen(p) + 1;
	}

done:
  free(reader_names);
	pcsc_started = TRUE;
	atexit(pcsc_cleanup);
	return;

failed:
  free(reader_names);
  exit(EXIT_FAILURE);
}

char *pcsc_reader_find(char *reader_name)
{
  int i;

  pcsc_startup();

	if (reader_name == NULL)
	{
    if (pcsc_readers_count)
		  return pcsc_readers_list[0];
		return NULL;
	}

	for (i=0; i<pcsc_readers_count; i++)
	{
#ifdef WIN32
		if (!stricmp(pcsc_readers_list[i], reader_name))
			return reader_name;
#else
		if (!strcasecmp(pcsc_readers_list[i], reader_name))
			return reader_name;
#endif
	}

	i = atoi(reader_name);

	if ((i > 0) && (i <= pcsc_readers_count))
	  return pcsc_readers_list[i - 1];

	return NULL;
}

void pcsc_cleanup(void)
{
  int i;

  if (pcsc_readers_list != NULL)
	{
    for (i=0; i<pcsc_readers_count; i++)
		  free(pcsc_readers_list[i]);
    free(pcsc_readers_list);
	}

  SCardReleaseContext(scardcontext);
}

BOOL pcsc_is_card_present(char *reader_name)
{
	SCARD_READERSTATE stState;
	LONG rc;

  stState.szReader = reader_name;
	stState.dwCurrentState = SCARD_STATE_UNAWARE;

	rc = SCardGetStatusChange(scardcontext, 0, &stState, 1);	
	if (rc != SCARD_S_SUCCESS)
  {
    printf("SCardGetStatusChange : %08lX\n",rc);
		goto failed;
  }

	if (stState.dwEventState & SCARD_STATE_PRESENT)
	  return TRUE;
	return FALSE;

failed:
  exit(EXIT_FAILURE);
	return FALSE;
}

BOOL pcsc_wait_card_present(char *reader_name)
{
	SCARD_READERSTATE stState;
	LONG rc;

  stState.szReader = reader_name;
	stState.dwCurrentState = SCARD_STATE_UNAWARE;

again:

	rc = SCardGetStatusChange(scardcontext, INFINITE, &stState, 1);	
	if (rc != SCARD_S_SUCCESS)
  {
    printf("SCardGetStatusChange : %08lX\n",rc);
		goto failed;
  }


	if (stState.dwEventState & SCARD_STATE_PRESENT)
	{
	  return TRUE;
	}

	stState.dwCurrentState = stState.dwEventState;
	goto again;

failed:
  exit(EXIT_FAILURE);
	return FALSE;
}

BOOL pcsc_wait_card_absent(char *reader_name)
{
	SCARD_READERSTATE stState;
	LONG rc;

  stState.szReader = reader_name;
	stState.dwCurrentState = SCARD_STATE_UNAWARE;

again:

	rc = SCardGetStatusChange(scardcontext, INFINITE, &stState, 1);	
	if (rc != SCARD_S_SUCCESS)
  {
    printf("SCardGetStatusChange : %08lX\n",rc);
		goto failed;
  }


	if (!(stState.dwEventState & SCARD_STATE_PRESENT))
	{
	  return TRUE;
	}

	stState.dwCurrentState = stState.dwEventState;
	goto again;

failed:
  exit(EXIT_FAILURE);
	return FALSE;
}


struct _smartcard_st
{
	SCARDHANDLE hCard;
  char *reader;
  BOOL direct;
	unsigned long protocol;
  unsigned char atr[32];
	size_t atrlen;
};

smartcard_t pcsc_connect(char *reader_name)
{
  LONG rc;
	unsigned long dummy_1, dummy_2, atrlen;
  struct _smartcard_st *card = malloc(sizeof(struct _smartcard_st));

	if (card == NULL)
	{
	  perror("malloc");
    exit(EXIT_FAILURE);
	}

  card->direct = FALSE;
	card->reader = reader_name;

  rc = SCardConnect(scardcontext,
                    card->reader,
                    SCARD_SHARE_SHARED,
                    SCARD_PROTOCOL_T0|SCARD_PROTOCOL_T1,
                    &card->hCard,
                    &card->protocol);
  if(rc != SCARD_S_SUCCESS)
  {
	  if (!be_silent)
      printf("SCardConnect : %08lX\n",rc);
    free(card);
	  return NULL;
  }

  atrlen = sizeof(card->atr);
	rc = SCardStatus(card->hCard, NULL, NULL, &dummy_1, &dummy_2, card->atr, &atrlen);
  if(rc != SCARD_S_SUCCESS)
  {
	  if (!be_silent)
      printf("SCardStatus : %08lX\n",rc);
		SCardDisconnect(card->hCard, SCARD_EJECT_CARD);
    free(card);
	  return NULL;
  }
	card->atrlen = (unsigned char) atrlen;

  return card;
}

smartcard_t pcsc_connect_direct(char *reader_name)
{
  LONG rc;
  struct _smartcard_st *card = malloc(sizeof(struct _smartcard_st));

	if (card == NULL)
	{
	  perror("malloc");
    exit(EXIT_FAILURE);
	}

  card->direct = TRUE;
	card->reader = reader_name;

  rc = SCardConnect(scardcontext,
                    card->reader,
                    SCARD_SHARE_DIRECT,
                    0,
                    &card->hCard,
                    &card->protocol);
  if(rc != SCARD_S_SUCCESS)
  {
	  if (!be_silent)
      printf("SCardConnect (direct) : %08lX\n",rc);
    free(card);
	  return NULL;
  }

  return card;
}

BOOL pcsc_disconnect(smartcard_t card)
{
  LONG rc;

  if (card == NULL)
    return FALSE;
	
  rc = SCardDisconnect(card->hCard, SCARD_EJECT_CARD);

	free(card);

  if(rc != SCARD_S_SUCCESS)
  {
	  if (!be_silent)
      printf("SCardDisconnect : %08lX\n",rc);
    return FALSE;
  }

  return TRUE;
}

SCARDHANDLE smartcard_handle(smartcard_t card)
{
  if (card == NULL)
	  return (SCARDHANDLE) 0;
  return card->hCard;
}

BOOL pcsc_transmit(smartcard_t card, const unsigned char capdu[], unsigned long capdu_length, unsigned char rapdu[], unsigned long *rapdu_length)
{
  LONG rc;
	const SCARD_IO_REQUEST *pioSendPci;

  if ((card == NULL) || (capdu == NULL) || (rapdu == NULL) || (rapdu_length == NULL))
	  return FALSE;

	switch (card->protocol)
	{
	  case SCARD_PROTOCOL_T0 : pioSendPci = SCARD_PCI_T0; break;
    case SCARD_PROTOCOL_T1 : pioSendPci = SCARD_PCI_T1; break;
		default                : pioSendPci = NULL;
	}

  rc = SCardTransmit(card->hCard,
	                   pioSendPci,
										 capdu,
										 capdu_length,
										 NULL,
										 rapdu,
										 rapdu_length);

  if (rc != SCARD_S_SUCCESS)
	{
	  if (!be_silent)
      printf("SCardTransmit error %08lX\n", rc);
		return FALSE;
	}

  if (*rapdu_length < 2)
	{
    if (!be_silent)
    {
      unsigned long i;
		  printf("Invalid response from reader (R-APDU without SW)\n");
      printf("\tC : ");
      for (i=0; i<capdu_length; i++)
        printf("%02X", capdu[i]);
      printf("\n");
      printf("\tR : ");
      for (i=0; i<*rapdu_length; i++)
        printf("%02X", rapdu[i]);
      printf("\n");
    }
    return FALSE;
	}

  return TRUE;
}

BOOL pcsc_control(smartcard_t card, const unsigned char capdu[], unsigned long capdu_length, unsigned char rapdu[], unsigned long *rapdu_length)
{
  LONG rc;
  unsigned long rapdu_size;

  if ((card == NULL) || (capdu == NULL) || (rapdu == NULL) || (rapdu_length == NULL))
	  return FALSE;

  rapdu_size = *rapdu_length;
  *rapdu_length = 0;

  rc = SCardControl(card->hCard,
	                  IOCTL_VENDOR_ESCAPE_SPRINGCARD,
                    capdu,
										capdu_length,
									  rapdu,
										rapdu_size,
                    rapdu_length);

  if (rc != SCARD_S_SUCCESS)
	{
	  if (!be_silent)
      printf("SCardControl error %08lX\n", rc);
		return FALSE;
	}

  if (*rapdu_length < 1)
  {
    if (be_verbose)
      printf("Invalid response from reader (control R-APDU empty)\n");
    return FALSE;
  }

  return TRUE;
}

